#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <Servo.h>

#include "konfiguracja.h"

#ifdef ESP32
  #pragma message(Program tylko dla ukladow typu ESP8266!)
  #error Wybierz uklad ESP8266.
#endif
// ----------------------------------------------------------------------------------
// ---  STALE I ZMIENNE
// ----------------------------------------------------------------------------------  
//----- KONFIGURACJA SERW --------
  Servo SERWO_A;
  int SERWO_A_POS = SERWO_NEUTRUM; // neutrum - polozenie srodkowe
  float SERWO_A_KOR_F = (float)SERWO_A_ZAKRES / (float)SERWO_MAX;  
  // Servo SERWO_B;

//----- KONFIGURACJA SILNIKA -----
  float MOTOR_KOR_F = (float)MOTOR_PWM_MAX_MIN / (float)MOTOR_PWM_WE_DOST;

//----- KONFIGURACJA WIFI --------
  WiFiUDP UDP;
  IPAddress local_IP(192,168,2,1);
  IPAddress gateway(192,168,2,10);
  IPAddress subnet(255,255,255,0);
  char BUFOR_PAKIETU_UDP[255];                 

//----- ZABEZPIECZENIE ------------
//      UTRATY ZASIEGU
  int WIFI_WATCHDOG = 0;
  const int ZWLOKA_CZASOWA = 150; // ms
  int WATCHDOG_BLOKADA = 1;

//----- POMIAR NAPIECIA AKU -------
  int ADC_X10_LICZ = 0;
  int ADC_X10_OPOZNIENIE = 0;
  int ADC_POMIAR = 0;
  int NAPIECIE_ZMIERZONE_X10 = 0;
  int PRZEKROCZENIE_NAPIECIA = 0;
  int ALARM_BATERI = 0;

//--- FUNKCJE KONFIGURACYJNE ------
  void IO_SET(void);
  void RS232_SET(void);
  void SERWO_SET(void);
  int WIFI_SET(void);
  
//------- FUNKCJE WLASNE ----------
  inline void WDT_W_ZASIEGU(int CZAS_ZWLOKI);
  inline void LEDZ_ON(void);
  inline void LEDZ_OFF(void);
  inline void LEDC_ON(void);
  inline void LEDC_OFF(void);
  inline void USTAW_SERWO_A(int SERWO_POS);
  inline void AKTYWACJA_IO(void);
  inline void DEAKTYWACJA_IO(void);
  inline void MOTOR_STOP(void);
  inline void MOTOR_NAPRZOD(int PWM);
  inline void MOTOR_WSTECZ(int PWM);

// ----------------------------------------------------------------------------------
// ---  SETUP
// ----------------------------------------------------------------------------------  
void setup() {
  IO_SET();                                   // konfiguracja wyprowadzen
  analogWriteFreq(ESP8266_PWM_CZESTOTLIWOSC); // konfiguracja sygnału PWM dla silnika
  RS232_SET();                                // konfiguracja RS232
  SERWO_SET();                                // konfiguracja serw  
  
  if(WIFI_SET()==0){                          // konfiguracja WiFi
    UDP.begin(UDP_PORT);                      // Start serwera UDP
    LEDC_ON();
    LEDZ_ON();
  } else {
    #if DEBUG_WIFI_EN
      Serial.println("Blad podczas konfiguracji interfejsu sieciowego !");
    #endif    
    while(1){ // blokada dzialania programu    
      LEDC_ON();
      delay(250);
      LEDC_OFF();
      delay(250);
      ESP.wdtFeed();
    }
  }
} // KONIEC PROCEDURY SETUP !

// ----------------------------------------------------------------------------------
// ---  LOOP
// ----------------------------------------------------------------------------------
void loop() {
// put your main code here, to run repeatedly:
  #if KALIBRACJA_ADC_EN
      int WARTOSC_ADC = analogRead(A0);
      Serial.print("Kod napiecia: ");
      Serial.print(WARTOSC_ADC);
      Serial.print(", Napiecie [V]: ");
      Serial.println((WARTOSC_ADC * BAT_MON_DIV * WSP_KALIBRACJI)/10230);      
  #endif
  
  WDT_W_ZASIEGU(ZWLOKA_CZASOWA);  // Zabezpieczenie przy utracie zasiegu
// ----------------------------------------------------------------------------------
// ---  Odbior pakietu UDP
// ----------------------------------------------------------------------------------
  int URZADZENIA_POLACZONE = WiFi.softAPgetStationNum();
    if(URZADZENIA_POLACZONE > 0){ // ktos jest do nas podlaczony
      int ROZMIAR_PAKIETU_UDP = UDP.parsePacket();
      if (ROZMIAR_PAKIETU_UDP > 0) { // pakiet nie zerowej dlugosci
        
        #if DEBUG_UDP_EN
          Serial.print("Odebrano pakiet o rozmiarze : ");
          Serial.print(ROZMIAR_PAKIETU_UDP);
          Serial.print(" od ");
          IPAddress ZDALNE_IP = UDP.remoteIP();
          Serial.print(ZDALNE_IP);
          Serial.print(", port ");
          Serial.println(UDP.remotePort());
        #endif
        
        // odczyt pakietu do bufora
        int DLUGOSC_PAKIETU_UDP = UDP.read(BUFOR_PAKIETU_UDP, 255);

        #if DEBUG_UDP_EN
          if (DLUGOSC_PAKIETU_UDP > 0) {
            BUFOR_PAKIETU_UDP[DLUGOSC_PAKIETU_UDP] = 0; // w zasadzie ma sens tylko wtedy jesli chcemy czytac bufor jako string
          }
          Serial.print("Zawartosc: ");
          Serial.println(BUFOR_PAKIETU_UDP);
        #endif

        // -----------------------------------------------------------------------------
        // -------   Pomiar napiecia akumulatora
        // -----------------------------------------------------------------------------
        #if BAT_MON_EN
          const int INT_BAT_VALARM_X10 = (int)BAT_VALARM_X10;         
          if (ADC_X10_OPOZNIENIE >= 10) { // pomiar co 10 pakietów ~1s
            if (ADC_X10_LICZ >= 10) {                        
              if (NAPIECIE_ZMIERZONE_X10 < INT_BAT_VALARM_X10) { // napiecie zmierzone ponizej progu alarmowego
                if(PRZEKROCZENIE_NAPIECIA < 4) PRZEKROCZENIE_NAPIECIA++;
              } else {
                if(ALARM_BATERI==0){
                  PRZEKROCZENIE_NAPIECIA = 0;
                }
              }                          
              NAPIECIE_ZMIERZONE_X10 = 0;
              ADC_X10_LICZ = 0;
            } else {
              ADC_X10_LICZ++;
              ADC_POMIAR = analogRead(A0); // pomiar i napiecia z adc
              NAPIECIE_ZMIERZONE_X10 += ADC_POMIAR;
            }
            ADC_X10_OPOZNIENIE = 1;      
          } else {
            ADC_X10_OPOZNIENIE++;
          }

          // Przekroczenie min. poziomu akumulatora przez ~15s. Wylaczenie silnika i serwa
          if(PRZEKROCZENIE_NAPIECIA > 3){
            // SERWO_A
            USTAW_SERWO_A(SERWO_NEUTRUM); // neutrum
            // SILNIK  
            MOTOR_STOP();
            DEAKTYWACJA_IO();
            ALARM_BATERI = 1;     
          }
          
          #if DEBUG_VMON_EN
            Serial.print("Kod napiecia: ");
            Serial.print(INT_BAT_VALARM_X10);
            Serial.print(" Pomiar napiecia x10 [kod]: ");
            Serial.print(NAPIECIE_ZMIERZONE_X10);
            Serial.print(" PRZEKROCZENIE_NAPIECIA [licznik]: ");
            Serial.print(PRZEKROCZENIE_NAPIECIA);            
            Serial.print(" ALARM_BATERI : ");
            Serial.print(ALARM_BATERI);
            Serial.print(" Pomiar napiecia ADC [kod]: ");
            Serial.print(ADC_POMIAR);            
            Serial.print(" Pomiar napiecia ADC [V]: ");
            Serial.println((ADC_POMIAR * BAT_MON_DIV * WSP_KALIBRACJI)/1024);
          #endif
          
        #endif
                
        // -----------------------------------------------------------------------------
        // -------   Sterowanie Serwem A
        // -----------------------------------------------------------------------------
        int POZYCJA_SERWA_A = (BUFOR_PAKIETU_UDP[5]-48);
        POZYCJA_SERWA_A += (BUFOR_PAKIETU_UDP[4]-48) * 10;
        POZYCJA_SERWA_A += (BUFOR_PAKIETU_UDP[3]-48) * 100;
        POZYCJA_SERWA_A += (BUFOR_PAKIETU_UDP[2]-48) * 1000;
        POZYCJA_SERWA_A -= SERWO_NEUTRUM; // roznica wzgledem neutrum

        // odwracanie kierunku obrotu
        #if SERWO_A_REWERS
          POZYCJA_SERWA_A = -POZYCJA_SERWA_A;
        #endif
        
        float POZYCJA_SERWA_A_KOREKTA = POZYCJA_SERWA_A * SERWO_A_KOR_F;
        
        SERWO_A_POS = SERWO_A_TRYM + SERWO_NEUTRUM + (int) POZYCJA_SERWA_A_KOREKTA;
        
        USTAW_SERWO_A(SERWO_A_POS);

        #if DEBUG_SERWO_EN
          Serial.print("Serwo A : ");
          Serial.print(POZYCJA_SERWA_A); Serial.print(" ; ");
          Serial.print(POZYCJA_SERWA_A_KOREKTA); Serial.print(" ; ");
          Serial.print(SERWO_A_TRYM);  Serial.print(" ; ");
          Serial.println(SERWO_A_POS);
        #endif
        
        // -----------------------------------------------------------------------------
        // -------   Sterowanie Silnikiem
        // -----------------------------------------------------------------------------
        int PREDKOSC_SILNIKA = (BUFOR_PAKIETU_UDP[11]-48);
        PREDKOSC_SILNIKA += (BUFOR_PAKIETU_UDP[10]-48) * 10;
        PREDKOSC_SILNIKA += (BUFOR_PAKIETU_UDP[9]-48) * 100;
        PREDKOSC_SILNIKA += (BUFOR_PAKIETU_UDP[8]-48) * 1000;
        
        PREDKOSC_SILNIKA -= MOTOR_ZERO; // roznica wzgledem neutrum, (ujema do tyłu / dodatnia do przodu)

        if(PREDKOSC_SILNIKA < 0){
          PREDKOSC_SILNIKA = - PREDKOSC_SILNIKA;
          PREDKOSC_SILNIKA -= MOTOR_LUZ;
          MOTOR_NAPRZOD(PREDKOSC_SILNIKA);          
        } else {
          PREDKOSC_SILNIKA -= MOTOR_LUZ;
          MOTOR_WSTECZ(PREDKOSC_SILNIKA);
        }
        
        // Reset WATCHDOGa WiFi
        WIFI_WATCHDOG = 0;
        WATCHDOG_BLOKADA = 0;                       
      }      
    }  
      
} // KONIEC PETLI LOOP !

// ----------------------------------------------------------------------------------
// --- Funkcje konfiguracyjne
// ----------------------------------------------------------------------------------
void IO_SET(void){
  pinMode(WYBUF_ENN_PIN, OUTPUT);     digitalWrite(WYBUF_ENN_PIN,1);
  pinMode(LEDC_CTRL_PIN, OUTPUT);     digitalWrite(LEDC_CTRL_PIN,1);
  pinMode(LEDZ_CTRL_PIN, OUTPUT);     digitalWrite(LEDZ_CTRL_PIN,1);
  pinMode(TRYM_KIER_PIN, INPUT);  // finalnie nie uzywany
  pinMode(SILNIK_CTRL_A_PIN, OUTPUT); digitalWrite(SILNIK_CTRL_A_PIN,0);
  pinMode(SILNIK_CTRL_B_PIN, OUTPUT); digitalWrite(SILNIK_CTRL_B_PIN,0);
  pinMode(SBUF_EN_PIN, OUTPUT);       digitalWrite(SBUF_EN_PIN,0);
  pinMode(SERWO_A_PIN, OUTPUT);       digitalWrite(SERWO_A_PIN,0);
  pinMode(SERWO_B_PIN, OUTPUT);       digitalWrite(SERWO_B_PIN,0);  
}

void RS232_SET(void){
  #if DEBUG_EN
    Serial.begin(RS232_PREDKOSC);
    Serial.println();
  #endif  
}

void SERWO_SET(void){
  SERWO_A.attach(SERWO_A_PIN);
  SERWO_A.writeMicroseconds(SERWO_A_POS);
  // SERWO_B.attach(SERWO_B_PIN);  
}

int WIFI_SET(void){
  int powodzenie = 0;
  
  WiFi.mode(WIFI_AP);
  WiFi.disconnect();
  delay(100);

  // Konfiguracja programowego punktu dostepowego
  #if DEBUG_WIFI_EN
    Serial.print("Konfiguracja Soft-AP ... ");
  #endif
  if (WiFi.softAPConfig(local_IP, gateway, subnet) != true){
    #if DEBUG_WIFI_EN
      Serial.println("Usterka");
    #endif
    powodzenie = 1;
  } else {
    #if DEBUG_WIFI_EN
      Serial.println("Gotowy");
    #endif
  }
  
  // Start punktu dostepowego
  #if DEBUG_WIFI_EN
    Serial.print("Soft-AP Start ... ");
  #endif
  
  if(WiFi.softAP(WIFI_AP_SSID, WIFI_AP_PASS, WIFI_AP_CHANNEL, WIFI_AP_HIDE) != true){
    #if DEBUG_WIFI_EN
      Serial.println("Usterka");
    #endif
    powodzenie = 1;
   } else {
    #if DEBUG_WIFI_EN
      Serial.println("Gotowy");
    #endif
   }
   return(powodzenie);
}

// ----------------------------------------------------------------------------------
// --- Funkcje wlasne
// ----------------------------------------------------------------------------------
inline void WDT_W_ZASIEGU(int CZAS_ZWLOKI){
    
  unsigned long NOWE_MILISEKUNDY;
  static unsigned long STARE_MILISEKUNDY;
  
  NOWE_MILISEKUNDY = millis();
  if(WATCHDOG_BLOKADA == 0){
    WIFI_WATCHDOG += (int) (NOWE_MILISEKUNDY - STARE_MILISEKUNDY);
    if(ALARM_BATERI==0){
      AKTYWACJA_IO();
      LEDZ_ON();
      LEDC_OFF();
    }else{
      LEDZ_OFF();
      LEDC_ON();      
    }
  }
  STARE_MILISEKUNDY = NOWE_MILISEKUNDY;
  
  // reset sterowania do polozen neutralnych  
  if(WIFI_WATCHDOG > CZAS_ZWLOKI){
    // SERWO_A
    USTAW_SERWO_A(SERWO_NEUTRUM); // neutrum
    // SILNIK  
    MOTOR_STOP();
    LEDC_ON();
    WATCHDOG_BLOKADA = 1;
  }  
}

inline void LEDZ_ON(void){
  digitalWrite(LEDZ_CTRL_PIN,0);
}

inline void LEDZ_OFF(void){
  digitalWrite(LEDZ_CTRL_PIN,1);
}

inline void LEDC_ON(void){
  digitalWrite(LEDC_CTRL_PIN,0);
}

inline void LEDC_OFF(void){
  digitalWrite(LEDC_CTRL_PIN,1);
}

inline void USTAW_SERWO_A(int SERWO_POS){
  SERWO_A.writeMicroseconds(SERWO_POS);
}

inline void AKTYWACJA_IO(void){
  digitalWrite(WYBUF_ENN_PIN,0);
}

inline void DEAKTYWACJA_IO(void){
  digitalWrite(WYBUF_ENN_PIN,1);
}

inline void MOTOR_STOP(void){
  digitalWrite(SILNIK_CTRL_A_PIN,0);
  digitalWrite(SILNIK_CTRL_B_PIN,0);
  analogWrite(SBUF_EN_PIN, ESP8266_PWM_MAX);  //1023
}

inline void MOTOR_NAPRZOD(int PWM){
  if(PWM > 0){
    float PWM_MOD = MOTOR_KOR_F * PWM;
    PWM_MOD += MOTOR_PWM_MIN;
    analogWrite(SBUF_EN_PIN, (int)PWM_MOD);
    digitalWrite(SILNIK_CTRL_A_PIN,1);
    digitalWrite(SILNIK_CTRL_B_PIN,0); 
          
    #if DEBUG_MOTOR_EN
      Serial.print("MOTOR_NAPRZOD : ");
      Serial.println((int)PWM_MOD);
    #endif        
          
  } else {
    MOTOR_STOP();
  }
}

inline void MOTOR_WSTECZ(int PWM){
  if(PWM > 0){
    float PWM_MOD = MOTOR_KOR_F * PWM;
    PWM_MOD += MOTOR_PWM_MIN;
    analogWrite(SBUF_EN_PIN, (int)PWM_MOD);
    digitalWrite(SILNIK_CTRL_A_PIN,0);
    digitalWrite(SILNIK_CTRL_B_PIN,1);
          
    #if DEBUG_MOTOR_EN
      Serial.print("MOTOR_WSTECZ : ");
      Serial.println((int)PWM_MOD);
    #endif      
          
  } else {
    MOTOR_STOP();
  }
}
